"""
The `~certbot_dns_desec.dns_desec` plugin automates the process of
completing a ``dns-01`` challenge (`~acme.challenges.DNS01`) by creating, and
subsequently removing, TXT records using the deSEC REST API.


Named Arguments
---------------

========================================  =====================================
``--dns-desec-credentials``               deSEC Remote API credentials_
                                          INI file. (Required)
``--dns-desec-propagation-seconds``       The number of seconds to wait for DNS
                                          to propagate before asking the ACME
                                          server to verify the DNS record.
                                          (Default: 120)  # TODO default, needed?
========================================  =====================================


Credentials
-----------

Use of this plugin requires a configuration file containing a deSEC DNS API
access token obtained from the deSEC website.

.. code-block:: ini
   :name: credentials.ini
   :caption: Example credentials file:

   # deSEC API token used by Certbot
   dns_desec_token = asdf
   dns_desec_endpoint = https://localhost:8080

The path to this file can be provided interactively or using the
``--dns-desec-credentials`` command-line argument. Certbot records the path
to this file for use during renewal, but does not store the file's contents.

.. caution::
   You should protect these API credentials as you would a password. Users who
   can read this file can use these credentials to issue API calls on
   your behalf. Users who can cause Certbot to run using these credentials can
   complete a ``dns-01`` challenge to acquire new certificates or revoke
   existing certificates for associated domains, even if those domains aren't
   being managed by this server.

Certbot will emit a warning if it detects that the credentials file can be
accessed by other users on your system. The warning reads "Unsafe permissions
on credentials configuration file", followed by the path to the credentials
file. This warning will be emitted each time Certbot uses the credentials file,
including for renewal, and cannot be silenced except by addressing the issue
(e.g., by using a command like ``chmod 600`` to restrict access to the file).

Examples
--------

.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``

   certbot certonly \\
     --dns-desec \\
     --dns-desec-credentials ~/.secrets/certbot/desec.ini \\
     -d example.com

.. code-block:: bash
   :caption: To acquire a single certificate for both ``example.com`` and
             ``www.example.com``

   certbot certonly \\
     --dns-desec \\
     --dns-desec-credentials ~/.secrets/certbot/desec.ini \\
     -d example.com \\
     -d www.example.com

.. code-block:: bash
   :caption: To acquire a certificate for ``example.com``, waiting 240 seconds
             for DNS propagation

   certbot certonly \\
     --dns-desec \\
     --dns-desec-credentials ~/.secrets/certbot/desec.ini \\
     --dns-desec-propagation-seconds 240 \\  # TODO
     -d example.com

"""



"""DNS Authenticator for deSEC."""
import json
import logging

import requests
import zope.interface
from certbot import errors
from certbot import interfaces
from certbot.plugins import dns_common

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@zope.interface.implementer(interfaces.IAuthenticator)
@zope.interface.provider(interfaces.IPluginFactory)
class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for deSEC

    This Authenticator uses the deSEC REST API to fulfill a dns-01 challenge.
    """

    description = "Obtain certificates using a DNS TXT record (if you are using deSEC.io for DNS)."
    ttl = 60

    def __init__(self, *args, **kwargs):
        super(Authenticator, self).__init__(*args, **kwargs)
        self.credentials = None

    @classmethod
    def add_parser_arguments(cls, add):  # pylint: disable=arguments-differ
        super(Authenticator, cls).add_parser_arguments(
            add, default_propagation_seconds=5  # TODO
        )
        add("credentials", help="deSEC credentials INI file.")

    def more_info(self):  # pylint: disable=missing-docstring,no-self-use
        return (
            "This plugin configures a DNS TXT record to respond to a dns-01 challenge using "
            "the deSEC Remote REST API."
        )

    def _setup_credentials(self):
        self.credentials = self._configure_credentials(
            "credentials",
            "deSEC credentials INI file",
            {
                # "endpoint": "URL of the deSEC API.", # TODO add documentation for this
                "token": "Access token for deSEC API.",
            },
        )

    def _perform(self, domain, validation_name, validation):
        logger.debug(f"Authenticator._perform: {domain}, {validation_name}, {validation}")
        subname = validation_name.split(domain)[0].rstrip('.')  # TODO code duplication
        records = self._get_desec_client().get_txt_rrset(domain, subname)
        self._get_desec_client().set_txt_rrset(domain, subname, list(records | {f'"{validation}"'}), self.ttl)

    def _cleanup(self, domain, validation_name, validation):
        logger.debug(f"Authenticator._cleanup: {domain}, {validation_name}, {validation}")
        subname = validation_name.split(domain)[0].rstrip('.')
        records = self._get_desec_client().get_txt_rrset(domain, subname)
        self._get_desec_client().set_txt_rrset(domain, subname, list(records - {f'"{validation}"'}), self.ttl)

    def _get_desec_client(self):
        return _DesecConfigClient(
            self.credentials.conf("endpoint") or "https://desec.io/api/v1/",
            self.credentials.conf("token"),
        )


class _DesecConfigClient(object):
    """
    Encapsulates all communication with the deSEC REST API.
    """

    def __init__(self, endpoint, token):
        logger.debug("creating _DesecConfigClient")
        self.endpoint = endpoint
        self.token = token
        self.session = requests.Session()
        self.session.headers["Authorization"] = f"Token {token}"
        self.session.headers["Content-Type"] = "application/json"

    def get_txt_rrset(self, domain, subname):
        response = self.session.get(
            url=f"{self.endpoint}/domains/{domain}/rrsets/{subname}/TXT",
        )

        if response.status_code == 404:
            return set()

        self._check_response_status(response, domain)
        return set(self._response_json(response).get('records', set()))

    def set_txt_rrset(self, domain, subname, rrset, ttl):
        response = self.session.put(
            url=f"{self.endpoint}/domains/{domain}/rrsets/",
            data=json.dumps([
                {"subname": subname, "type": "TXT", "ttl": ttl, "records": rrset},
            ]),
        )
        return self._check_response_status(response, domain)

    def _check_response_status(self, response, domain):
        if 200 <= response.status_code <= 299:
            return
        elif response.status_code in [401, 403]:
            raise errors.PluginError(f"Could not authenticate against deSEC API: {response.content}")
        elif response.status_code == 404:
            raise errors.PluginError(f"Could not find domain '{domain}': {response.content}")
        elif response.status_code == 429:
            raise errors.PluginError(f"deSEC throttled your request. Please run certbot for various domains at "
                                     f"different times. {response.content}")
        elif response.status_code >= 500:
            raise errors.PluginError(f"deSEC API server error (status {response.status_code}): {response.content}")
        else:
            raise errors.PluginError(f"Unknown error when talking to deSEC (status {response.status_code}: "
                                     f"Request was on '{response.request.url}' with payload {response.request.body}. "
                                     f"Response was '{response.content}'.")

    def _response_json(self, response):
        try:
            return response.json()
        except json.JSONDecodeError:
            raise errors.PluginError(f"deSEC API sent non-JSON response (status {response.status_code}): "
                                     f"{response.content}")



from setuptools import setup
from setuptools import find_packages

version = "0.1.2"

install_requires = [
    "acme>=0.29.0",
    "certbot>=0.34.0",
    "setuptools",
    "requests",
    "mock",
    "requests-mock",
]

# read the contents of your README file
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst")) as f:
    long_description = f.read()

setup(
    name="certbot-dns-desec",
    version=version,
    description="deSEC DNS Authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/desec-io/certbot-dns-desec",
    author="Nils Wisiol",
    author_email="nils@desec.io",
    license="Apache License 2.0",
    python_requires=">=3.7",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        "certbot.plugins": [
            "dns-desec = certbot_dns_desec.dns_desec:Authenticator"
        ]
    },
    test_suite="certbot_dns_desec",
)



"""Tests for certbot_dns_desec.dns_desec."""

import unittest

import mock
import json
import requests_mock

from certbot import errors
from certbot.compat import os
from certbot.plugins import dns_test_common
from certbot.plugins.dns_test_common import DOMAIN
from certbot.tests import util as test_util

FAKE_TOKEN = "faketoken"
FAKE_ENDPOINT = "mock://endpoint"


class AuthenticatorTest(
    test_util.TempDirTestCase, dns_test_common.BaseAuthenticatorTest
):
    def setUp(self):
        super(AuthenticatorTest, self).setUp()

        from certbot_dns_desec.dns_desec import Authenticator

        path = os.path.join(self.tempdir, "file.ini")
        dns_test_common.write(
            {
                "desec_token": FAKE_TOKEN,
                "desec_endpoint": FAKE_ENDPOINT,
            },
            path,
        )

        super(AuthenticatorTest, self).setUp()
        self.config = mock.MagicMock(
            desec_credentials=path, desec_propagation_seconds=0
        )  # don't wait during tests

        self.auth = Authenticator(self.config, "desec")

        self.mock_client = mock.MagicMock()
        # _get_desec_client | pylint: disable=protected-access
        self.auth._get_desec_client = mock.MagicMock(return_value=self.mock_client)

    def test_perform(self):
        self.auth.perform([self.achall])

        expected = [
            mock.call.get_txt_rrset(DOMAIN, "_acme-challenge"),
            mock.ANY, mock.ANY, mock.ANY, mock.ANY,  # TODO make nice
            mock.call.set_txt_rrset(DOMAIN, "_acme-challenge", mock.ANY, mock.ANY),
        ]
        self.assertEqual(expected, self.mock_client.mock_calls)

    def test_cleanup(self):
        # _attempt_cleanup | pylint: disable=protected-access
        self.auth._attempt_cleanup = True
        self.auth.cleanup([self.achall])

        expected = [
            mock.call.get_txt_rrset(DOMAIN, "_acme-challenge"),
            mock.ANY, mock.ANY, mock.ANY, mock.ANY,  # TODO make nice
            mock.call.set_txt_rrset(DOMAIN, "_acme-challenge", mock.ANY, mock.ANY),
        ]
        self.assertEqual(expected, self.mock_client.mock_calls)


class DesecConfigClientTest(unittest.TestCase):
    record_name = "foo"
    record_content = "bar"
    record_ttl = 42

    def setUp(self):
        from certbot_dns_desec.dns_desec import _DesecConfigClient

        self.adapter = requests_mock.Adapter()

        self.client = _DesecConfigClient(FAKE_ENDPOINT, FAKE_TOKEN)
        self.client.session.mount("mock", self.adapter)

    def _register_response(self, url, response=None, requires_token=True, status=200):
        def additional_matcher(request):
            okay = True
            okay &= request.headers["Content-Type"] == "application/json"
            if requires_token:
                okay &= request.headers["Authorization"] == f"Token {FAKE_TOKEN}"
            return okay

        self.adapter.register_uri(
            method=requests_mock.ANY,
            url=url,
            text=json.dumps(response),
            additional_matcher=additional_matcher,
            status_code=status,
        )

    def test_set_txt_rrset(self):
        self._register_response(
            url=f"{FAKE_ENDPOINT}/domains/{DOMAIN}/rrsets/",
            response=[
                {
                    "created": "2021-05-13T09:38:41.576975Z",
                    "domain": DOMAIN,
                    "subname": "_acme_challenge",
                    "name": self.record_name,
                    "records": [self.record_content],
                    "ttl": self.record_ttl,
                    "type": "TXT",
                    "touched": "2021-05-13T09:38:41.585257Z"
                }
            ]
        )

        self.client.set_txt_rrset(
            DOMAIN, self.record_name, self.record_content, self.record_ttl
        )

    def test_set_txt_rrset_fail_to_find_domain(self):
        self._register_response(
            url=f"{FAKE_ENDPOINT}/domains/{DOMAIN}/rrsets/",
            response={"detail": "Not found."},
            status=404,
        )
        with self.assertRaises(errors.PluginError):
            self.client.set_txt_rrset(
                DOMAIN, self.record_name, self.record_content, self.record_ttl
            )

    def test_set_txt_rrset_fail_to_authenticate(self):
        self._register_response(
            url=f"{FAKE_ENDPOINT}/domains/{DOMAIN}/rrsets/",
            response={"detail": "Invalid token."},
            status=403,
        )
        with self.assertRaises(errors.PluginError):
            self.client.set_txt_rrset(
                DOMAIN, self.record_name, self.record_content, self.record_ttl
            )


if __name__ == "__main__":
    unittest.main()  # pragma: no cover
